import abc
from random import Random

from .mypy_types import OptInt


class SamplerABC(abc.ABC):
    @abc.abstractmethod
    def is_sampled(self, trace_id: str) -> bool:  # pragma: no cover
        """Defines if given trace should be recorded for further actions."""
        pass


class Sampler(SamplerABC):
    def __init__(self, *, sample_rate: float = 1.0, seed: OptInt = None) -> None:
        self._sample_rate = sample_rate
        self._rng = Random(seed)

    def is_sampled(self, trace_id: str) -> bool:
        if self._sample_rate == 0.0:
            sampled = False
        else:
            sampled = self._rng.random() <= self._sample_rate
        return sampled


# TODO: implement other types of sampler for example hash trace_id



import time
from typing import Any, Dict, List, NamedTuple, Optional

from .mypy_types import Headers, OptBool, OptInt, OptStr, OptTs


# possible span kinds
CLIENT = "CLIENT"
SERVER = "SERVER"
PRODUCER = "PRODUCER"
CONSUMER = "CONSUMER"

# zipkin headers, for more information see:
# https://github.com/openzipkin/b3-propagation

TRACE_ID_HEADER = "X-B3-TraceId"
SPAN_ID_HEADER = "X-B3-SpanId"
PARENT_ID_HEADER = "X-B3-ParentSpanId"
FLAGS_HEADER = "X-B3-Flags"
SAMPLED_ID_HEADER = "X-B3-Sampled"
SINGLE_HEADER = "b3"
DELIMITER = "-"
DEBUG_MARKER = "d"


class _TraceContext(NamedTuple):
    trace_id: str
    parent_id: OptStr
    span_id: str
    sampled: OptBool
    debug: bool
    shared: bool


class TraceContext(_TraceContext):
    """Immutable class with trace related data that travels across
    process boundaries.
    """

    def make_headers(self) -> Headers:
        """Creates dict with zipkin headers from available context.

        Resulting dict should be passed to HTTP client  propagate contest
        to other services.
        """
        return make_headers(self)

    def make_single_header(self) -> Headers:
        return make_single_header(self)


class Endpoint(NamedTuple):
    serviceName: OptStr
    ipv4: OptStr
    ipv6: OptStr
    port: OptInt


def create_endpoint(
    service_name: str, *, ipv4: OptStr = None, ipv6: OptStr = None, port: OptInt = None
) -> Endpoint:
    """Factory function to create Endpoint object."""
    return Endpoint(service_name, ipv4, ipv6, port)


def make_timestamp(ts: OptTs = None) -> int:
    """Create zipkin timestamp in microseconds, or convert available one
    from second. Useful when user supplies ts from time.time() call.
    """
    ts = ts if ts is not None else time.time()
    return int(ts * 1000 * 1000)  # microseconds


def make_headers(context: TraceContext) -> Headers:
    """Creates dict with zipkin headers from supplied trace context."""
    headers = {
        TRACE_ID_HEADER: context.trace_id,
        SPAN_ID_HEADER: context.span_id,
        FLAGS_HEADER: "0",
        SAMPLED_ID_HEADER: "1" if context.sampled else "0",
    }
    if context.parent_id is not None:
        headers[PARENT_ID_HEADER] = context.parent_id
    return headers


def make_single_header(context: TraceContext) -> Headers:
    """Creates dict with zipkin single header format."""
    # b3={TraceId}-{SpanId}-{SamplingState}-{ParentSpanId}
    c = context

    # encode sampled flag
    if c.debug:
        sampled = "d"
    elif c.sampled:
        sampled = "1"
    else:
        sampled = "0"

    params: List[str] = [c.trace_id, c.span_id, sampled]
    if c.parent_id is not None:
        params.append(c.parent_id)

    h = DELIMITER.join(params)
    headers = {SINGLE_HEADER: h}
    return headers


def parse_sampled_header(headers: Headers) -> OptBool:
    sampled = headers.get(SAMPLED_ID_HEADER.lower(), None)
    if sampled is None or sampled == "":
        return None
    return True if sampled == "1" else False


def parse_debug_header(headers: Headers) -> bool:
    return True if headers.get(FLAGS_HEADER, "0") == "1" else False


def _parse_parent_id(parts: List[str]) -> OptStr:
    # parse parent_id part from zipkin single header propagation
    parent_id = None
    if len(parts) >= 4:
        parent_id = parts[3]
    return parent_id


def _parse_debug(parts: List[str]) -> bool:
    # parse debug part from zipkin single header propagation
    debug = False
    if len(parts) >= 3 and parts[2] == DEBUG_MARKER:
        debug = True
    return debug


def _parse_sampled(parts: List[str]) -> OptBool:
    # parse sampled part from zipkin single header propagation
    sampled: OptBool = None
    if len(parts) >= 3:
        if parts[2] in ("1", "0"):
            sampled = bool(int(parts[2]))
    return sampled


def _parse_single_header(headers: Headers) -> Optional[TraceContext]:
    # Makes TraceContext from zipkin single header format.
    # https://github.com/openzipkin/b3-propagation

    # b3={TraceId}-{SpanId}-{SamplingState}-{ParentSpanId}
    if headers[SINGLE_HEADER] == "0":
        return None
    payload = headers[SINGLE_HEADER].lower()
    parts: List[str] = payload.split(DELIMITER)
    if len(parts) < 2:
        return None

    debug = _parse_debug(parts)
    sampled = debug if debug else _parse_sampled(parts)

    context = TraceContext(
        trace_id=parts[0],
        span_id=parts[1],
        parent_id=_parse_parent_id(parts),
        sampled=sampled,
        debug=debug,
        shared=False,
    )
    return context


def make_context(headers: Headers) -> Optional[TraceContext]:
    """Converts available headers to TraceContext, if headers mapping does
    not contain zipkin headers, function returns None.
    """
    # TODO: add validation for trace_id/span_id/parent_id

    # normalize header names just in case someone passed regular dict
    # instead dict with case insensitive keys
    headers = {k.lower(): v for k, v in headers.items()}

    required = (TRACE_ID_HEADER.lower(), SPAN_ID_HEADER.lower())
    has_b3 = all(h in headers for h in required)
    has_b3_single = SINGLE_HEADER in headers

    if not (has_b3_single or has_b3):
        return None

    if has_b3:
        debug = parse_debug_header(headers)
        sampled = debug if debug else parse_sampled_header(headers)
        context = TraceContext(
            trace_id=headers[TRACE_ID_HEADER.lower()],
            parent_id=headers.get(PARENT_ID_HEADER.lower()),
            span_id=headers[SPAN_ID_HEADER.lower()],
            sampled=sampled,
            debug=debug,
            shared=False,
        )
        return context
    return _parse_single_header(headers)


OptKeys = Optional[List[str]]


def filter_none(data: Dict[str, Any], keys: OptKeys = None) -> Dict[str, Any]:
    """Filter keys from dict with None values.

    Check occurs only on root level. If list of keys specified, filter
    works only for selected keys
    """

    def limited_filter(k: str, v: Any) -> bool:
        return k not in keys or v is not None  # type: ignore

    def full_filter(k: str, v: Any) -> bool:
        return v is not None

    f = limited_filter if keys is not None else full_filter
    return {k: v for k, v in data.items() if f(k, v)}



import abc
import asyncio
import warnings
from collections import deque
from typing import Any, Awaitable, Callable, Deque, Dict, List, Optional, Tuple

import aiohttp
from aiohttp.client_exceptions import ClientError
from yarl import URL

from .log import logger
from .mypy_types import OptLoop
from .record import Record


DEFAULT_TIMEOUT = aiohttp.ClientTimeout(total=5 * 60)
BATCHES_MAX_COUNT = 10 ** 4

DataList = List[Dict[str, Any]]
SndBatches = Deque[Tuple[int, DataList]]
SendDataCoro = Callable[[DataList], Awaitable[bool]]


class TransportABC(abc.ABC):
    @abc.abstractmethod
    def send(self, record: Record) -> None:  # pragma: no cover
        """Sends data to zipkin collector."""
        pass

    @abc.abstractmethod
    async def close(self) -> None:  # pragma: no cover
        """Performs additional cleanup actions if required."""
        pass


class StubTransport(TransportABC):
    """Dummy transport, which logs spans to a limited queue."""

    def __init__(self, queue_length: int = 100) -> None:
        logger.info("Zipkin address was not provided, using stub transport")
        self.records: Deque[Record] = deque(maxlen=queue_length)

    def send(self, record: Record) -> None:
        self.records.append(record)

    async def close(self) -> None:
        pass


class BatchManager:
    def __init__(
        self,
        max_size: int,
        send_interval: float,
        attempt_count: int,
        send_data: SendDataCoro,
    ) -> None:
        loop = asyncio.get_event_loop()
        self._max_size = max_size
        self._send_interval = send_interval
        self._send_data = send_data
        self._attempt_count = attempt_count
        self._max = BATCHES_MAX_COUNT
        self._sending_batches: SndBatches = deque([], maxlen=self._max)
        self._active_batch: Optional[DataList] = None
        self._ender = loop.create_future()
        self._timer: Optional[asyncio.Future[Any]] = None
        self._sender_task = asyncio.ensure_future(self._sender_loop())

    def add(self, data: Dict[str, Any]) -> None:
        if self._active_batch is None:
            self._active_batch = []
        self._active_batch.append(data)
        if len(self._active_batch) >= self._max_size:
            self._sending_batches.append((0, self._active_batch))
            self._active_batch = None
            if self._timer is not None and not self._timer.done():
                self._timer.cancel()

    async def stop(self) -> None:
        self._ender.set_result(None)

        await self._sender_task
        await self._send()

        if self._timer is not None:
            self._timer.cancel()
            try:
                await self._timer
            except asyncio.CancelledError:
                pass

    async def _sender_loop(self) -> None:
        while not self._ender.done():
            await self._wait()
            await self._send()

    async def _send(self) -> None:
        if self._active_batch is not None:
            self._sending_batches.append((0, self._active_batch))
            self._active_batch = None

        batches = self._sending_batches.copy()
        self._sending_batches = deque([], maxlen=self._max)
        for attempt, batch in batches:
            if not await self._send_data(batch):
                attempt += 1
                if attempt < self._attempt_count:
                    self._sending_batches.append((attempt, batch))

    async def _wait(self) -> None:
        self._timer = asyncio.ensure_future(asyncio.sleep(self._send_interval))

        await asyncio.wait(
            [self._timer, self._ender],
            return_when=asyncio.FIRST_COMPLETED,
        )


class Transport(TransportABC):
    def __init__(
        self,
        address: str,
        send_interval: float = 5,
        loop: OptLoop = None,
        *,
        send_max_size: int = 100,
        send_attempt_count: int = 3,
        send_timeout: Optional[aiohttp.ClientTimeout] = None
    ) -> None:
        if loop is not None:
            warnings.warn(
                "loop parameter is deprecated and ignored",
                DeprecationWarning,
                stacklevel=2,
            )
        self._address = URL(address)
        self._queue: DataList = []
        self._closing = False
        self._send_interval = send_interval
        if send_timeout is None:
            send_timeout = DEFAULT_TIMEOUT
        self._session = aiohttp.ClientSession(
            timeout=send_timeout,
            headers={"Content-Type": "application/json"},
        )
        self._batch_manager = BatchManager(
            send_max_size,
            send_interval,
            send_attempt_count,
            self._send_data,
        )

    def send(self, record: Record) -> None:
        data = record.asdict()
        self._batch_manager.add(data)

    async def _send_data(self, data: DataList) -> bool:
        try:

            async with self._session.post(self._address, json=data) as resp:
                body = await resp.text()
                if resp.status >= 300:
                    msg = "zipkin responded with code: {} and body: {}".format(
                        resp.status, body
                    )
                    raise RuntimeError(msg)

        except (asyncio.TimeoutError, ClientError):
            return False
        except Exception as exc:  # pylint: disable=broad-except
            # that code should never fail and break application
            logger.error("Can not send spans to zipkin", exc_info=exc)
        return True

    async def close(self) -> None:
        if self._closing:
            return

        self._closing = True
        await self._batch_manager.stop()
        await self._session.close()



import ipaddress
import sys
from contextlib import contextmanager
from types import SimpleNamespace
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Generator,
    Iterable,
    Optional,
    Set,
    cast,
)

import aiohttp
from aiohttp import (
    TraceRequestEndParams,
    TraceRequestExceptionParams,
    TraceRequestStartParams,
)
from aiohttp.web import (
    AbstractRoute,
    Application,
    HTTPException,
    Request,
    StreamResponse,
    middleware,
)

from .constants import HTTP_METHOD, HTTP_PATH, HTTP_ROUTE, HTTP_STATUS_CODE
from .helpers import (
    CLIENT,
    SERVER,
    TraceContext,
    make_context,
    parse_debug_header,
    parse_sampled_header,
)
from .span import SpanAbc
from .tracer import Tracer


APP_AIOZIPKIN_KEY = "aiozipkin_tracer"
REQUEST_AIOZIPKIN_KEY = "aiozipkin_span"


__all__ = (
    "setup",
    "get_tracer",
    "request_span",
    "middleware_maker",
    "make_trace_config",
    "APP_AIOZIPKIN_KEY",
    "REQUEST_AIOZIPKIN_KEY",
)

Handler = Callable[[Request], Awaitable[StreamResponse]]
Middleware = Callable[[Request, Handler], Awaitable[StreamResponse]]


def _set_remote_endpoint(span: SpanAbc, request: Request) -> None:
    peername = request.remote
    if peername is not None:
        kwargs: Dict[str, Any] = {}
        try:
            peer_ipaddress = ipaddress.ip_address(peername)
        except ValueError:
            pass
        else:
            if isinstance(peer_ipaddress, ipaddress.IPv4Address):
                kwargs["ipv4"] = str(peer_ipaddress)
            else:
                kwargs["ipv6"] = str(peer_ipaddress)
        if kwargs:
            span.remote_endpoint(None, **kwargs)


def _get_span(request: Request, tracer: Tracer) -> SpanAbc:
    # builds span from incoming request, if no context found, create
    # new span
    context = make_context(request.headers)

    if context is None:
        sampled = parse_sampled_header(request.headers)
        debug = parse_debug_header(request.headers)
        span = tracer.new_trace(sampled=sampled, debug=debug)
    else:
        span = tracer.join_span(context)
    return span


def _set_span_properties(span: SpanAbc, request: Request) -> None:
    span_name = f"{request.method.upper()} {request.path}"
    span.name(span_name)
    span.kind(SERVER)
    span.tag(HTTP_PATH, request.path)
    span.tag(HTTP_METHOD, request.method.upper())

    resource = request.match_info.route.resource
    assert resource is not None
    route = resource.canonical
    span.tag(HTTP_ROUTE, route)

    _set_remote_endpoint(span, request)


PY37 = sys.version_info >= (3, 7)

if PY37:
    from contextvars import ContextVar

    OptTraceVar = ContextVar[Optional[TraceContext]]
    zipkin_context: OptTraceVar = ContextVar("zipkin_context", default=None)

    @contextmanager
    def set_context_value(
        context_var: OptTraceVar, value: TraceContext
    ) -> Generator[OptTraceVar, None, None]:
        token = context_var.set(value)
        try:
            yield context_var
        finally:
            context_var.reset(token)


def middleware_maker(
    skip_routes: Optional[Iterable[AbstractRoute]] = None,
    tracer_key: str = APP_AIOZIPKIN_KEY,
    request_key: str = REQUEST_AIOZIPKIN_KEY,
) -> Middleware:
    s = skip_routes
    skip_routes_set: Set[AbstractRoute] = set(s) if s else set()

    @middleware
    async def aiozipkin_middleware(
        request: Request, handler: Handler
    ) -> StreamResponse:
        # route is in skip list, we do not track anything with zipkin
        if request.match_info.route in skip_routes_set:
            resp = await handler(request)
            return resp

        tracer = request.app[tracer_key]
        span = _get_span(request, tracer)
        request[request_key] = span
        if span.is_noop:
            resp = await handler(request)
            return resp

        if PY37:
            with set_context_value(zipkin_context, span.context):
                with span:
                    _set_span_properties(span, request)
                    try:
                        resp = await handler(request)
                    except HTTPException as e:
                        span.tag(HTTP_STATUS_CODE, str(e.status))
                        raise
                    span.tag(HTTP_STATUS_CODE, str(resp.status))
        else:
            with span:
                _set_span_properties(span, request)
                try:
                    resp = await handler(request)
                except HTTPException as e:
                    span.tag(HTTP_STATUS_CODE, str(e.status))
                    raise
                span.tag(HTTP_STATUS_CODE, str(resp.status))

        return resp

    return aiozipkin_middleware


def setup(
    app: Application,
    tracer: Tracer,
    *,
    skip_routes: Optional[Iterable[AbstractRoute]] = None,
    tracer_key: str = APP_AIOZIPKIN_KEY,
    request_key: str = REQUEST_AIOZIPKIN_KEY,
) -> Application:
    """Sets required parameters in aiohttp applications for aiozipkin.

    Tracer added into application context and cleaned after application
    shutdown. You can provide custom tracer_key, if default name is not
    suitable.
    """
    app[tracer_key] = tracer
    m = middleware_maker(
        skip_routes=skip_routes, tracer_key=tracer_key, request_key=request_key
    )
    app.middlewares.append(m)

    # register cleanup signal to close zipkin transport connections
    async def close_aiozipkin(app: Application) -> None:
        await app[tracer_key].close()

    app.on_cleanup.append(close_aiozipkin)

    return app


def get_tracer(app: Application, tracer_key: str = APP_AIOZIPKIN_KEY) -> Tracer:
    """Returns tracer object from application context.

    By default tracer has APP_AIOZIPKIN_KEY in aiohttp application context,
    you can provide own key, if for some reason default one is not suitable.
    """
    return cast(Tracer, app[tracer_key])


def request_span(request: Request, request_key: str = REQUEST_AIOZIPKIN_KEY) -> SpanAbc:
    """Returns span created by middleware from request context, you can use it
    as parent on next child span.
    """
    return cast(SpanAbc, request[request_key])


class ZipkinClientSignals:
    """Class contains signal handler for aiohttp client. Handlers executed
    only if aiohttp session contains tracer context with span.
    """

    def __init__(self, tracer: Tracer) -> None:
        self._tracer = tracer

    def _get_span_context(
        self, trace_config_ctx: SimpleNamespace
    ) -> Optional[TraceContext]:
        ctx = self._get_span_context_from_dict(
            trace_config_ctx
        ) or self._get_span_context_from_namespace(trace_config_ctx)

        if ctx:
            return ctx

        if PY37:
            has_implicit_context = zipkin_context.get() is not None
            if has_implicit_context:
                return zipkin_context.get()

        return None

    def _get_span_context_from_dict(
        self, trace_config_ctx: SimpleNamespace
    ) -> Optional[TraceContext]:
        ctx = trace_config_ctx.trace_request_ctx

        if isinstance(ctx, dict):
            r: Optional[TraceContext] = ctx.get("span_context")
            return r

        return None

    def _get_span_context_from_namespace(
        self, trace_config_ctx: SimpleNamespace
    ) -> Optional[TraceContext]:
        ctx = trace_config_ctx.trace_request_ctx

        if isinstance(ctx, SimpleNamespace):
            r: Optional[TraceContext] = getattr(ctx, "span_context", None)
            return r

        return None

    async def on_request_start(
        self,
        session: aiohttp.ClientSession,
        context: SimpleNamespace,
        params: TraceRequestStartParams,
    ) -> None:
        span_context = self._get_span_context(context)
        if span_context is None:
            return

        p = params
        span = self._tracer.new_child(span_context)
        context._span = span
        span.start()
        span_name = f"client {p.method.upper()} {p.url.path}"
        span.name(span_name)
        span.kind(CLIENT)

        ctx = context.trace_request_ctx
        propagate_headers = True

        if isinstance(ctx, dict):
            # Check ctx is dict to be compatible with old package versions
            propagate_headers = ctx.get("propagate_headers", True)
        if isinstance(ctx, SimpleNamespace):
            propagate_headers = getattr(ctx, "propagate_headers", True)

        if propagate_headers:
            span_headers = span.context.make_headers()
            p.headers.update(span_headers)

    async def on_request_end(
        self,
        session: aiohttp.ClientSession,
        context: SimpleNamespace,
        params: TraceRequestEndParams,
    ) -> None:
        span_context = self._get_span_context(context)
        if span_context is None:
            return

        span = context._span
        span.finish()
        delattr(context, "_span")

    async def on_request_exception(
        self,
        session: aiohttp.ClientSession,
        context: SimpleNamespace,
        params: TraceRequestExceptionParams,
    ) -> None:

        span_context = self._get_span_context(context)
        if span_context is None:
            return
        span = context._span
        span.finish(exception=params.exception)
        delattr(context, "_span")


def make_trace_config(tracer: Tracer) -> aiohttp.TraceConfig:
    """Creates aiohttp.TraceConfig with enabled aiozipking instrumentation
    for aiohttp client.
    """
    trace_config = aiohttp.TraceConfig()
    zipkin = ZipkinClientSignals(tracer)

    trace_config.on_request_start.append(zipkin.on_request_start)
    trace_config.on_request_end.append(zipkin.on_request_end)
    trace_config.on_request_exception.append(zipkin.on_request_exception)
    return trace_config



import warnings
from typing import (  # noqa
    TYPE_CHECKING,
    Any,
    AsyncContextManager,
    Awaitable,
    Dict,
    List,
    Optional,
    Type,
)

from .context_managers import _ContextManager
from .helpers import Endpoint, TraceContext
from .mypy_types import OptBool, OptLoop
from .record import Record
from .sampler import Sampler, SamplerABC
from .span import NoopSpan, Span, SpanAbc
from .transport import StubTransport, Transport, TransportABC
from .utils import generate_random_64bit_string, generate_random_128bit_string


if TYPE_CHECKING:

    class _Base(AsyncContextManager["Tracer"]):
        pass


else:

    class _Base(AsyncContextManager):
        pass


class Tracer(_Base):
    def __init__(
        self,
        transport: TransportABC,
        sampler: SamplerABC,
        local_endpoint: Endpoint,
        ignored_exceptions: Optional[List[Type[Exception]]] = None,
    ) -> None:
        super().__init__()
        self._records: Dict[TraceContext, Record] = {}
        self._transport = transport
        self._sampler = sampler
        self._local_endpoint = local_endpoint
        self._ignored_exceptions = ignored_exceptions or []

    def new_trace(self, sampled: OptBool = None, debug: bool = False) -> SpanAbc:
        context = self._next_context(None, sampled=sampled, debug=debug)
        return self.to_span(context)

    def join_span(self, context: TraceContext) -> SpanAbc:
        new_context = context
        if context.sampled is None:
            sampled = self._sampler.is_sampled(context.trace_id)
            new_context = new_context._replace(sampled=sampled)
        else:
            new_context = new_context._replace(shared=True)
        return self.to_span(new_context)

    def new_child(self, context: TraceContext) -> SpanAbc:
        new_context = self._next_context(context)
        if not context.sampled:
            return NoopSpan(self, new_context, self._ignored_exceptions)
        return self.to_span(new_context)

    def to_span(self, context: TraceContext) -> SpanAbc:
        if not context.sampled:
            return NoopSpan(self, context, self._ignored_exceptions)

        record = Record(context, self._local_endpoint)
        self._records[context] = record
        return Span(self, context, record, self._ignored_exceptions)

    def _send(self, record: Record) -> None:
        self._records.pop(record.context, None)
        self._transport.send(record)

    def _next_context(
        self,
        context: Optional[TraceContext] = None,
        sampled: OptBool = None,
        debug: bool = False,
    ) -> TraceContext:
        span_id = generate_random_64bit_string()
        if context is not None:
            new_context = context._replace(
                span_id=span_id, parent_id=context.span_id, shared=False
            )
            return new_context

        trace_id = generate_random_128bit_string()
        if sampled is None:
            sampled = self._sampler.is_sampled(trace_id)

        new_context = TraceContext(
            trace_id=trace_id,
            parent_id=None,
            span_id=span_id,
            sampled=sampled,
            debug=debug,
            shared=False,
        )
        return new_context

    async def close(self) -> None:
        await self._transport.close()

    async def __aenter__(self) -> "Tracer":
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self.close()


def create(
    zipkin_address: str,
    local_endpoint: Endpoint,
    *,
    sample_rate: float = 0.01,
    send_interval: float = 5,
    loop: OptLoop = None,
    ignored_exceptions: Optional[List[Type[Exception]]] = None
) -> _ContextManager[Tracer]:
    if loop is not None:
        warnings.warn(
            "loop parameter is deprecated and ignored", DeprecationWarning, stacklevel=2
        )

    async def build_tracer() -> Tracer:
        sampler = Sampler(sample_rate=sample_rate)
        transport = Transport(zipkin_address, send_interval=send_interval)
        return Tracer(transport, sampler, local_endpoint, ignored_exceptions)

    result = _ContextManager(build_tracer())
    return result


def create_custom(
    local_endpoint: Endpoint,
    transport: Optional[TransportABC] = None,
    sampler: Optional[SamplerABC] = None,
    ignored_exceptions: Optional[List[Type[Exception]]] = None,
) -> _ContextManager[Tracer]:
    t = transport or StubTransport()
    sample_rate = 1  # sample everything
    s = sampler or Sampler(sample_rate=sample_rate)

    async def build_tracer() -> Tracer:
        return Tracer(t, s, local_endpoint, ignored_exceptions)

    result = _ContextManager(build_tracer())
    return result



import random
import struct
import time


# https://github.com/Yelp/py_zipkin/blob/
# 61f8aa3412f6c1b4e1218ed34cb117e97cc9a6cc/py_zipkin/util.py#L22-L75
def generate_random_64bit_string() -> str:
    """Returns a 64 bit UTF-8 encoded string. In the interests of simplicity,
    this is always cast to a `str` instead of (in py2 land) a unicode string.
    Certain clients (I'm looking at you, Twisted) don't enjoy unicode headers.

    :returns: random 16-character string
    """
    return "{:016x}".format(random.getrandbits(64))


# https://github.com/Yelp/py_zipkin/blob/
# 61f8aa3412f6c1b4e1218ed34cb117e97cc9a6cc/py_zipkin/util.py#L32
def generate_random_128bit_string() -> str:
    """Returns a 128 bit UTF-8 encoded string. Follows the same conventions
    as generate_random_64bit_string().

    The upper 32 bits are the current time in epoch seconds, and the
    lower 96 bits are random. This allows for AWS X-Ray `interop
    <https://github.com/openzipkin/zipkin/issues/1754>`_

    :returns: 32-character hex string
    """
    t = int(time.time())
    lower_96 = random.getrandbits(96)
    return "{:032x}".format((t << 96) | lower_96)


def unsigned_hex_to_signed_int(hex_string: str) -> int:
    """Converts a 64-bit hex string to a signed int value.

    This is due to the fact that Apache Thrift only has signed values.

    Examples:
        '17133d482ba4f605' => 1662740067609015813
        'b6dbb1c2b362bf51' => -5270423489115668655

    :param hex_string: the string representation of a zipkin ID
    :returns: signed int representation
    """
    v: int = struct.unpack("q", struct.pack("Q", int(hex_string, 16)))[0]
    return v


def signed_int_to_unsigned_hex(signed_int: int) -> str:
    """Converts a signed int value to a 64-bit hex string.

    Examples:
        1662740067609015813  => '17133d482ba4f605'
        -5270423489115668655 => 'b6dbb1c2b362bf51'

    :param signed_int: an int to convert
    :returns: unsigned hex string
    """
    hex_string = hex(struct.unpack("Q", struct.pack("q", signed_int))[0])[2:]
    if hex_string.endswith("L"):
        return hex_string[:-1]
    return hex_string



from .aiohttp_helpers import (
    APP_AIOZIPKIN_KEY,
    REQUEST_AIOZIPKIN_KEY,
    get_tracer,
    make_trace_config,
    middleware_maker,
    request_span,
    setup,
)
from .constants import (
    HTTP_HOST,
    HTTP_METHOD,
    HTTP_PATH,
    HTTP_REQUEST_SIZE,
    HTTP_RESPONSE_SIZE,
    HTTP_ROUTE,
    HTTP_STATUS_CODE,
    HTTP_URL,
)
from .helpers import CLIENT, CONSUMER, PRODUCER, SERVER, create_endpoint, make_context
from .sampler import Sampler
from .span import SpanAbc
from .tracer import Tracer, create, create_custom


__version__ = "1.1.0"
__all__ = (
    "Tracer",
    "Sampler",
    "SpanAbc",
    "create",
    "create_custom",
    "create_endpoint",
    "make_context",
    # aiohttp helpers
    "setup",
    "get_tracer",
    "request_span",
    "middleware_maker",
    "make_trace_config",
    "APP_AIOZIPKIN_KEY",
    "REQUEST_AIOZIPKIN_KEY",
    # possible span kinds
    "CLIENT",
    "SERVER",
    "PRODUCER",
    "CONSUMER",
    # constants
    "HTTP_HOST",
    "HTTP_METHOD",
    "HTTP_PATH",
    "HTTP_REQUEST_SIZE",
    "HTTP_RESPONSE_SIZE",
    "HTTP_STATUS_CODE",
    "HTTP_URL",
    "HTTP_ROUTE",
)



import os
import re
import sys

from setuptools import find_packages, setup  # type: ignore


if sys.version_info < (3, 6, 0):
    raise RuntimeError("aiozipkin does not support Python earlier than 3.6.0")


def read(f: str) -> str:
    return open(os.path.join(os.path.dirname(__file__), f)).read().strip()


install_requires = ["aiohttp>=3.7.2"]


def read_version() -> str:
    regexp = re.compile(r'^__version__\W*=\W*"([\d.abrc]+)"')
    init_py = os.path.join(os.path.dirname(__file__), "aiozipkin", "__init__.py")
    with open(init_py) as f:
        for line in f:
            match = regexp.match(line)
            if match is not None:
                return match.group(1)
        else:
            msg = "Cannot find version in aiozipkin/__init__.py"
            raise RuntimeError(msg)


classifiers = [
    "License :: OSI Approved :: MIT License",
    "Intended Audience :: Developers",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Operating System :: POSIX",
    "Development Status :: 5 - Production/Stable",
    "Framework :: AsyncIO",
]


setup(
    name="aiozipkin",
    version=read_version(),
    description=(
        "Distributed tracing instrumentation " "for asyncio application with zipkin"
    ),
    long_description="\n\n".join((read("README.rst"), read("CHANGES.rst"))),
    classifiers=classifiers,
    platforms=["POSIX"],
    author="Nikolay Novik",
    author_email="nickolainovik@gmail.com",
    url="https://github.com/aio-libs/aiozipkin",
    download_url="https://pypi.python.org/pypi/aiozipkin",
    license="Apache 2",
    packages=find_packages(),
    python_requires=">=3.6",
    install_requires=install_requires,
    keywords=["zipkin", "distributed-tracing", "tracing"],
    zip_safe=True,
    include_package_data=True,
)



from collections.abc import Awaitable as AbcAwaitable
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncContextManager,
    Awaitable,
    Generator,
    Generic,
    Optional,
    Type,
    TypeVar,
)


T = TypeVar("T", bound=AsyncContextManager["T"])  # type: ignore


if TYPE_CHECKING:

    class _Base(AsyncContextManager[T], AbcAwaitable[T]):
        pass


else:

    class _Base(Generic[T], AsyncContextManager, AbcAwaitable):
        pass


class _ContextManager(_Base[T]):

    __slots__ = ("_coro", "_obj")

    def __init__(self, coro: Awaitable[T]) -> None:
        super().__init__()
        self._coro: Awaitable[T] = coro
        self._obj: Optional[T] = None

    def __await__(self) -> Generator[Any, None, T]:
        return self._coro.__await__()

    async def __aenter__(self) -> T:
        self._obj = await self._coro
        t: T = await self._obj.__aenter__()
        return t

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> Optional[bool]:
        if self._obj is None:
            raise RuntimeError("__aexit__ called before __aenter__")
        return await self._obj.__aexit__(exc_type, exc, tb)
